import numpy as np

from rumorpheme.const import PAD, BEGIN, END, UNKNOWN


def prepare_data(data, symbol_codes, max_word_length=64):
    """
    Function to prepare data for training and inference.
    """
    batch_size = len(data)
    inputs = np.full((batch_size, max_word_length), PAD, dtype=int)
    inputs[:, 0] = BEGIN
    for i, word in enumerate(data):
        word_codes = [symbol_codes.get(char, UNKNOWN) for char in word]
        inputs[i, 1:1 + len(word_codes)] = word_codes
        inputs[i, 1 + len(word_codes)] = END
    return inputs



PAD, BEGIN, END, UNKNOWN = 0, 1, 2, 3

AUXILIARY = [
    'PAD',
    'BEGIN',
    'END',
    'UNKNOWN'
]

TARGET_SYMBOLS = [
    "PAD",
    "BEGIN",
    "END",
    "UNKNOWN",
    "-",
    "B-END",
    "B-POSTFIX",
    "B-PREF",
    "B-ROOT",
    "B-SUFF",
    "E-END",
    "E-POSTFIX",
    "E-PREF",
    "E-ROOT",
    "E-SUFF",
    "M-END",
    "M-POSTFIX",
    "M-PREF",
    "M-ROOT",
    "M-SUFF",
    "S-END",
    "S-HYPH",
    "S-LINK",
    "S-PREF",
    "S-ROOT",
    "S-SUFF"
]



import sys
import os
import json
import torch
import torch.nn as nn
from huggingface_hub import hf_hub_download

from .utils import prepare_data
from .const import TARGET_SYMBOLS


class RuMorphemeModel(nn.Module):
    def __init__(self, params, symbols, symbol_codes=None):
        super(RuMorphemeModel, self).__init__()

        # Available symbols
        self.symbols = symbols
        self.symbols_number = len(self.symbols)
        self.symbol_codes = symbol_codes
        if symbol_codes is None:
            self.symbol_codes = {a: i for i, a in enumerate(symbols)}

        # Target symbols
        self.target_symbols = TARGET_SYMBOLS
        self.target_symbols_number = len(self.target_symbols)
        self.target_symbol_codes = {a: i for i, a in enumerate(TARGET_SYMBOLS)}

        # Extract parameters from config
        self.use_morpheme_types = params.get('use_morpheme_types', True)
        self.measure_last = params.get("measure_last", self.use_morpheme_types)
        self.embeddings_size = params.get('embeddings_size', 32)
        self.conv_layers = params.get('conv_layers', 1)
        self.window_size = params.get('window_size', [5])
        self.filters_number = params.get('filters_number', 64)
        self.dense_output_units = params.get('dense_output_units', 0)
        self.use_lstm = params.get('use_lstm', False)
        self.lstm_units = params.get('lstm_units', 64)
        self.dropout = params.get('dropout', 0.0)
        self.context_dropout = params.get('context_dropout', 0.0)
        self.models_number = params.get('models_number', 1)

        # Ensure window_size and filters_number are lists
        if isinstance(self.window_size, int):
            self.window_size = [self.window_size]
        if isinstance(self.filters_number, int):
            self.filters_number = [self.filters_number] * len(self.window_size)

        # Define layers
        self.embedding = nn.Embedding(self.symbols_number, self.embeddings_size)
        self.conv_layers_list = nn.ModuleList()
        input_channels = self.embeddings_size
        for i in range(self.conv_layers):
            convs = nn.ModuleList()
            for ws, fn in zip(self.window_size, self.filters_number):
                conv = nn.Conv1d(input_channels, fn, ws, padding=ws // 2)
                convs.append(conv)
            self.conv_layers_list.append(convs)
            input_channels = sum(self.filters_number)
        if self.use_lstm:
            self.lstm = nn.LSTM(input_channels, self.lstm_units, batch_first=True, bidirectional=True)
            lstm_output_size = self.lstm_units * 2
            fc_input_dim = lstm_output_size
        else:
            fc_input_dim = input_channels
        if self.dense_output_units > 0:
            self.fc1 = nn.Linear(fc_input_dim, self.dense_output_units)
            self.fc_out = nn.Linear(self.dense_output_units, self.target_symbols_number)
        else:
            self.fc_out = nn.Linear(fc_input_dim, self.target_symbols_number)
        self.dropout_layer = nn.Dropout(self.dropout)

    def forward(self, x):
        """
        Function to perform forward pass of the model.
        :param x:
        :return:
        """
        x = self.embedding(x)  # [batch_size, seq_len, embedding_dim]
        x = x.permute(0, 2, 1)  # [batch_size, embedding_dim, seq_len]
        for convs in self.conv_layers_list:
            conv_outs = []
            for conv in convs:
                conv_out = conv(x)
                conv_out = torch.relu(conv_out)
                conv_outs.append(conv_out)
            x = torch.cat(conv_outs, dim=1)
            x = self.dropout_layer(x)
        x = x.permute(0, 2, 1)  # [batch_size, seq_len, channels]
        if self.use_lstm:
            x, _ = self.lstm(x)
        if self.dense_output_units > 0:
            x = torch.relu(self.fc1(x))
        logits = self.fc_out(x)  # [batch_size, seq_len, target_symbols_number]
        return logits

    @property
    def device(self):
        """Returns the device of the model."""
        return next(self.parameters()).device

    @classmethod
    def from_pretrained(cls, model_path, device=None):
        # Let's use cuda if available
        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        if os.path.isdir(model_path):
            # Load files from local directory
            config_file = os.path.join(model_path, "config.json")
            vocab_file = os.path.join(model_path, "vocab.json")
            model_file = os.path.join(model_path, "pytorch_model.bin")
        else:
            # Assume model_path is a Hugging Face repo ID
            repo_id = model_path
            config_file = hf_hub_download(repo_id=repo_id, filename="config.json")
            vocab_file = hf_hub_download(repo_id=repo_id, filename="vocab.json")
            model_file = hf_hub_download(repo_id=repo_id, filename="pytorch_model.bin")

        if not all(os.path.isfile(f) for f in [config_file, vocab_file, model_file]):
            print("Model files could not to be read.")
            sys.exit(1)

        # Read parameters of model
        with open(config_file, "r", encoding="utf8") as fin:
            params = json.load(fin)

        # Load vocabularies
        with open(vocab_file, "rb") as f:
            vocab_data = json.load(f)
        symbols = vocab_data["symbols"]
        symbol_codes = vocab_data["symbol_codes"]

        # Load model
        model = RuMorphemeModel(params, symbols, symbol_codes)
        model.load_state_dict(torch.load(model_file, map_location=device, weights_only=True))
        return model

    def predict(self, data: list):
        """
        Function to perform predictions of a word.
        :param data: it should be a list of words (or a list with a single word)
        :return:
        """
        # Disable gradient calculation
        with torch.no_grad():
            input_data = prepare_data(data, self.symbol_codes)
            inputs = torch.tensor(input_data, dtype=torch.long).to(self.device)  # Move data to same device as model
            outputs = self.forward(inputs)
            log_probs = torch.log_softmax(outputs, dim=-1)
            predictions = torch.argmax(log_probs, dim=-1)

        # Return predictions copied to CPU memory
        return predictions.cpu().numpy(), log_probs.cpu().numpy()



from setuptools import setup, find_packages

setup(
    name='ruMorpheme',
    version='0.2.1',
    author='Pavel Rykov',
    author_email='paul@drteam.rocks',
    description='Проект языковой модели для проведения морфемного анализа и сегментации слов русского языка.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/EvilFreelancer/ruMorpheme',
    packages=find_packages(),
    keywords='natural language processing, nlp, morpheme, russian',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing :: Linguistic',
        'Natural Language :: Russian',
    ],
    python_requires='>=3.11',
    install_requires=[
        "numpy~=2.1.1",
        "torch~=2.4.1",
        "huggingface-hub~=0.25.1",
    ],
)



import os
import json
import re
from typing import List

from tokenizers import pre_tokenizers, decoders, NormalizedString, PreTokenizedString, AddedToken
from transformers import PreTrainedTokenizerFast

from rumorpheme import RuMorphemeModel, labels_to_morphemes

DEFAULT_MODEL_NAME = "evilfreelancer/ruMorpheme-v0.2"

END, BEGIN, PAD, UNKNOWN, CAP, ALL_CAPS = 0, 1, 2, 3, 4, 5
SYSTEM, USER, ASSISTANT, FUNCTION_CALL, FUNCTION_RESPONSE = 6, 7, 8, 9, 10
SPACE = 11

AUXILIARY = [
    "</s>", "<s>", "<pad>", "<unk>", "<cap>", "<all_caps>",
    "system", "user", "assistant", "function_call", "function_response",
    " ",
]

NUMBERS = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]


class RuMorphemePreTokenizer:
    """
    Pre-tokenizer for RuMorpheme model.
    Splits on spaces and includes spaces as tokens.
    Then, applies morpheme splitting to non-space tokens.
    """

    def __init__(self, model_name: str = DEFAULT_MODEL_NAME):
        self.model = RuMorphemeModel.from_pretrained(model_name)
        self.model.eval()

    def pre_tokenize(self, pretok: PreTokenizedString):
        # First, split on spaces and include spaces as tokens
        pretok.split(self.split_on_spaces)
        # Then, apply morpheme splitting to non-space tokens
        pretok.split(self.morpheme_split)

    def split_on_spaces(self, i: int, normalized_string: NormalizedString) -> List[NormalizedString]:
        """
        Splits on spaces and includes spaces as tokens.
        TODO: Need to make performance tests on this function.
        """
        text = str(normalized_string)
        splits = [NormalizedString(match.group()) for match in re.finditer(r'\s+|\S+', text)]
        return splits

    def morpheme_split(self, i: int, normalized_string: NormalizedString) -> List[NormalizedString]:
        """
        Split word on morphemes, including numbers and punctuation.
        """
        word = str(normalized_string)

        # If word is just spaces or digits, return as is
        if word.isspace() or word.isdigit():
            return [normalized_string]

        # Ignore special characters (non-alphabetical)
        if not any(c.isalpha() for c in word):
            return [normalized_string]

        # Detect capitalization
        cap_token = None
        if word[0].isupper():
            cap_token = NormalizedString(AUXILIARY[CAP])
            if len(word) > 1 and word.isupper():
                cap_token = NormalizedString(AUXILIARY[ALL_CAPS])

        # Convert word to lowercase for morpheme splitting
        word_lower = word.lower()

        # Make predictions and return morphemes
        all_predictions, all_log_probs = self.model.predict([word_lower])
        morphs, morph_types, _ = labels_to_morphemes(word_lower, all_predictions[0], all_log_probs[0])

        # Create list of morpheme tokens
        morpheme_tokens = [
            NormalizedString(f"{morph_type}/{morph}")
            for morph, morph_type in zip(morphs, morph_types)
        ]

        # Insert capitalization token if needed
        if cap_token:
            return [cap_token] + morpheme_tokens
        else:
            return morpheme_tokens


class RuMorphemeDecoder:
    """
    Custom decoder for RuMorpheme model, it removes morph_type prefix from tokens and keeps spaces.
    """

    def decode_chain(self, tokens: List[str]) -> List[str]:
        """
        tokenizer.decode function calls this function
        """
        decoded_tokens = []
        capitalize_next = False
        uppercase_next = False

        for token in tokens:
            # Handle capitalization tokens
            if token == AUXILIARY[CAP]:
                capitalize_next = True
                continue
            elif token == AUXILIARY[ALL_CAPS]:
                uppercase_next = True
                continue

            # If token is a space, keep it as is
            if token.isspace():
                decoded_tokens.append(token)
            else:
                # Remove morph_type prefix if present
                if '/' in token:
                    _, morph = token.split('/', 1)
                else:
                    morph = token

                # Apply capitalization if needed
                if uppercase_next:
                    morph = morph.upper()
                    uppercase_next = False
                elif capitalize_next:
                    morph = morph.capitalize()
                    capitalize_next = False

                decoded_tokens.append(morph)
        return decoded_tokens


class RuMorphemeTokenizerFast(PreTrainedTokenizerFast):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # If pre-tokenizer nodel is not specified, use the default
        self.model_name = kwargs.get('model_name')
        if kwargs.get('model_name') is None:
            self.model_name: str = DEFAULT_MODEL_NAME

        # Complete initialization
        self.init_backend_tokenizer()

    def init_backend_tokenizer(self):
        # Custom pre-tokenizer
        self.backend_tokenizer.pre_tokenizer = pre_tokenizers.Sequence([
            pre_tokenizers.Punctuation(),
            pre_tokenizers.Digits(individual_digits=True),
            pre_tokenizers.PreTokenizer.custom(RuMorphemePreTokenizer(self.model_name))
        ])
        # Custom decoder
        self.backend_tokenizer.decoder = decoders.Decoder.custom(RuMorphemeDecoder())

    def save_pretrained(self, save_directory, **kwargs):
        # Temporarily remove the custom pre-tokenizer and decoder before saving
        original_pre_tokenizer = self.backend_tokenizer.pre_tokenizer
        original_decoder = self.backend_tokenizer.decoder
        self.backend_tokenizer.pre_tokenizer = None
        self.backend_tokenizer.decoder = None

        # Save the tokenizer using the parent method
        super().save_pretrained(save_directory, **kwargs)

        # Re-attach the custom pre-tokenizer and decoder
        self.backend_tokenizer.pre_tokenizer = original_pre_tokenizer
        self.backend_tokenizer.decoder = original_decoder

        # Save the tokenizer class name in tokenizer_config.json
        tokenizer_config_file = os.path.join(save_directory, 'tokenizer_config.json')
        if os.path.isfile(tokenizer_config_file):
            with open(tokenizer_config_file, 'r', encoding='utf-8') as f:
                tokenizer_config = json.load(f)
        else:
            tokenizer_config = {}

        # Correctly specify the tokenizer_class with module name
        tokenizer_config['tokenizer_class'] = "RuMorphemeTokenizerFast"
        tokenizer_config['use_fast'] = True
        tokenizer_config['auto_map'] = {"AutoTokenizer": ["", "tokenizer.RuMorphemeTokenizerFast"]}

        with open(tokenizer_config_file, 'w', encoding='utf-8') as f:
            json.dump(tokenizer_config, f, ensure_ascii=False)

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *init_inputs, **kwargs):
        # Load the tokenizer using the parent method
        tokenizer = super(RuMorphemeTokenizerFast, cls).from_pretrained(
            pretrained_model_name_or_path, *init_inputs, **kwargs
        )

        # If pre-tokenizer nodel is not specified, use the default
        model_name = kwargs.get('model_name')
        if kwargs.get('model_name') is None:
            model_name: str = DEFAULT_MODEL_NAME

        # Custom pre-tokenizer
        tokenizer.backend_tokenizer.pre_tokenizer = pre_tokenizers.Sequence([
            pre_tokenizers.Punctuation(),
            pre_tokenizers.Digits(individual_digits=True),
            pre_tokenizers.PreTokenizer.custom(RuMorphemePreTokenizer(model_name))
        ])

        # Custom decoder
        tokenizer.backend_tokenizer.decoder = decoders.Decoder.custom(RuMorphemeDecoder())

        return tokenizer



import numpy as np

from rumorpheme.const import TARGET_SYMBOLS


def labels_to_morphemes(word: str, labels, log_probs, use_morpheme_types=True):
    """
    Function for labeling predicted labels due to target symbols.
    :param word:
    :param labels:
    :param log_probs:
    :param use_morpheme_types:
    :return:
    """
    morphemes = []
    morpheme_types = []
    morpheme_log_probs = []
    curr_morpheme = ""
    curr_morpheme_log_probs = []
    prev_label_type = None  # Initialize prev_label_type
    if use_morpheme_types:
        end_labels = ['E-ROOT', 'E-PREF', 'E-SUFF', 'E-END', 'E-POSTFIX',
                      'S-ROOT', 'S-PREF', 'S-SUFF', 'S-END', 'S-LINK', 'S-HYPH']
    else:
        end_labels = ['E-None', 'S-None']
    for i, (letter, label_idx) in enumerate(zip(word, labels)):
        label = TARGET_SYMBOLS[label_idx]
        morpheme_type = label.split('-')[-1] if '-' in label else label
        if letter == '-':
            # Save current morpheme if any
            if curr_morpheme:
                morphemes.append(curr_morpheme)
                morpheme_types.append(prev_label_type if prev_label_type else 'UNKNOWN')
                avg_log_prob = sum(curr_morpheme_log_probs) / len(curr_morpheme_log_probs)
                morpheme_log_probs.append(avg_log_prob)
                curr_morpheme = ""
                curr_morpheme_log_probs = []
            # Treat hyphen as a separate morpheme of type HYPH
            morphemes.append(letter)
            morpheme_types.append('HYPH')
            morpheme_log_probs.append(log_probs[i][label_idx])
            prev_label_type = 'HYPH'
        else:
            curr_morpheme += letter
            curr_morpheme_log_probs.append(log_probs[i][label_idx])
            if label in end_labels:
                morphemes.append(curr_morpheme)
                morpheme_types.append(morpheme_type)
                avg_log_prob = sum(curr_morpheme_log_probs) / len(curr_morpheme_log_probs)
                morpheme_log_probs.append(avg_log_prob)
                curr_morpheme = ""
                curr_morpheme_log_probs = []
                prev_label_type = morpheme_type

    # Process any remaining morpheme
    if curr_morpheme:
        morphemes.append(curr_morpheme)
        morpheme_types.append(prev_label_type if prev_label_type else 'UNKNOWN')
        avg_log_prob = sum(curr_morpheme_log_probs) / len(curr_morpheme_log_probs)
        morpheme_log_probs.append(avg_log_prob)
    morpheme_probs = [np.exp(lp) * 100 for lp in morpheme_log_probs]
    return morphemes, morpheme_types, morpheme_probs
